/* vi:set ts=4 sw=4:

  textbuf.c    gapped buffer for ruby

*/

#include <stdio.h>
#ifdef __STDC__
# include <stdlib.h>
#endif

#include "ruby.h"
#include "re.h"

static int dodebug = 0;

/* declear ------------------------------------------------------ */

/*
    === struct textbuf ===

                  gap
                   |
                   V
    ptr -> OoOoOoOo--------------OoOoOoOoOoOoOo
                   |<- gaplen ->|
           |<------------ len --------------->|

    O : first byte of multi byte char
    o : rest byte
    - : gap


    gap never divide multi byte char
    nl is /\n|\r\n|\r/
    nl is one char
    index of first line is 1
    byte/char index of buffer-mark DOES NOT include gap length

*/

struct textbuf {
    char *ptr;              /* buffer */
    long len;               /* capacity of buffer (byte) */
    long gap;               /* byte index of gap */
    long gaplen;            /* byte length of gap */

    struct bufmark *marks;  /* d-linked-list of mark */
    struct bufmark *point;  /* edit point (internal use only) */
    struct bufmark *head;   /* head of string */
    struct bufmark *tail;   /* tail of string */
    VALUE vhead;            /* head mark */
    VALUE vtail;            /* tail mark */
};

struct bufmark {
    unsigned long flags;
    struct textbuf *buf;
    struct bufmark *prev;
    struct bufmark *next;
    long idx;
    long cidx;
    long line;
};

static VALUE TextBuffer;
static VALUE TextBufferMark;
static VALUE StringBuffer;
static VALUE StringBufferMark;


#ifdef DEBUG
static void p_buf _((struct textbuf *buf));
static void p_mark _((struct bufmark *mark));
#endif

/* creation */
static struct textbuf *buf_new _((unsigned long type));
static void buf_free _((struct textbuf *buf));
static struct bufmark *mark_new _((struct textbuf *buf,
                                   long idx, long cidx, long line,
                                   unsigned long flags));
static void mark_free _((struct bufmark *mark));

#define mark_new_default(buf, idx, cidx, line, flags) \
    mark_new(buf, idx, cidx, line, flags | INSERT_BEFORE)


/* mark (list) */
static void insert_mark_between _((struct bufmark *mark,
                                struct bufmark *prev, struct bufmark *next));
static void remove_mark _((struct bufmark *mark));
static struct bufmark *first_mark _((struct bufmark *mark));
static void adjust_mark_forward _((struct bufmark *mark));
static void adjust_mark_backward _((struct bufmark *mark));

/* mark (other) */
static void mark_forward_char _((struct bufmark *mark, long clen));
static void mark_forward_byte _((struct bufmark *mark, long len));
static void mark_back_char _((struct bufmark *mark, long clen));
static void mark_back_byte _((struct bufmark *mark, long len));
static void mark_move _((struct bufmark *mark, long len));
static void mark_nextline _((struct bufmark *mark));
static void mark_prevline _((struct bufmark *mark));
static void mark_linetail _((struct bufmark *mark));
static void mark_linehead _((struct bufmark *mark));
static long mark_col _((struct bufmark *mark));
static long mark_bcol _((struct bufmark *mark));
static void mark_insert _((struct bufmark *mark, char *str, long len));
static void mark_delete _((struct bufmark *mark, long len));

/* smark */
static void smark_move _((struct bufmark *mark, long len));
static void smark_insert _((struct bufmark *mark, char *str, long len));
static void smark_delete _((struct bufmark *mark, long len));

#define TEXTBUFMARK_FLAG  (1UL << 0)  /* mark for TextBuffer */
#define STRBUFMARK_FLAG   0UL         /* mark for StringBuffer */
#define TEXTBUFMARK_P(m)  (m->flags & TEXTBUFMARK_FLAG)
#define MARKTYPE(m)       TEXTBUFMARK_P(m)

#define ALIVE_FLAG        (1UL << 1)  /* mark->buf is not freed */
#define ALIVE_P(m)        (m->flags & ALIVE_FLAG)

#define INSERT_BEFORE     (1UL << 2)  /* insert text in "left" of mark */
#define INSERT_AFTER      0UL         /* insert text in "right" of mark */


/* buffer */
#if 0
static long searchi _((struct textbuf *buf, long beg,
                       int argtype, long arg, long *ch, long *li));
#endif
static long fwdnl _((struct textbuf *buf, long idx));
static long backnl _((struct textbuf *buf, long idx));
static long linehead _((struct textbuf *buf, long idx));
static long linetail _((struct textbuf *buf, long idx));
static long forward_nline _((struct textbuf *buf, long idx, long n));
static long back_nline _((struct textbuf *buf, long idx, long n));
static long byte_idx _((struct textbuf *buf,
                        long begidx, long clen, long *line));
static long byte_length _((struct textbuf *buf,
                           long begidx, long clen, long *line));
static long char_length _((struct textbuf *buf,
                           long begidx, long len, long *line));
static long char_idx _((struct textbuf *buf,
                        long begidx, long len, long *line));
static long fwdcol _((struct textbuf *buf, long beg, long clen));
static void setgap _((struct textbuf *buf, long idx));
static void insert _((struct textbuf *buf, long idx, char *str, long len));
static void delete _((struct textbuf *buf, long idx, long len));
static VALUE substr _((struct textbuf *buf, long idx, long len));
static VALUE getline _((struct textbuf *buf, long idx));
static void each_line_ro _((struct textbuf *buf, long idx));
static long legal_cidx _((struct textbuf *buf, long c));
static long legal_clen _((struct textbuf *buf, long cidx, long clen));
static long legal_idx _((struct textbuf *buf, long i));
static long legal_len _((struct textbuf *buf, long idx, long len));

#define INITLEN 4

#define LEN(buf) ((buf)->len - (buf)->gaplen)
#define CLEN(buf) ((buf)->tail->cidx)

#define LEN_STOP  (1 << 0)
#define CLEN_STOP (1 << 1)
#define LINE_STOP (1 << 2)
#define NO_LINE   (1 << 3)



/*
==================================================================
                             util
==================================================================
*/

#if defined(DEBUG) || defined(DEBUGB) || defined(DEBUGC) || \
    defined(DEBUGD) || defined(DEBUGI) || defined(DEBUGS)

static void
p_buf(buf)
    struct textbuf *buf;
{
    printf("#<textbuf 0x%lx len=%ld LEN=%ld gap=%ld gaplen=%ld>\n",
           (unsigned long)buf, buf->len, LEN(buf), buf->gap, buf->gaplen);
}

static void
p_mark(mark)
    struct bufmark *mark;
{
    printf("#<bufmark buf=0x%lx idx=%ld cidx=%ld>\n",
           (unsigned long)mark->buf, mark->idx, mark->cidx);
}

#endif

#ifdef DEBUG
# define D(code) code
#else
# define D(code)
#endif

#ifdef DEBUGB
# define Db(code) if (dodebug) code
#else
# define Db(code)
#endif

#ifdef DEBUGC
# define Dc(code) if (dodebug) code
#else
# define Dc(code)
#endif

#ifdef DEBUGD
# define Dd(code) if (dodebug) code
#else
# define Dd(code)
#endif

#ifdef DEBUGI
# define Di(code) if (dodebug) code
#else
# define Di(code)
#endif

#ifdef DEBUGS
# define Ds(code) if (dodebug) code
#else
# define Ds(code)
#endif



/*
==================================================================
                          creation
==================================================================
*/

static struct textbuf*
buf_new(type)
    unsigned long type;
{
    struct textbuf *buf;

    buf = ALLOC(struct textbuf);
    buf->ptr = 0;
    buf->len = 0;
    buf->gap = 0;
    buf->point = 0;

    buf->ptr = ALLOC_N(char, INITLEN);
    buf->ptr = calloc(sizeof(char), INITLEN);
/* printf("ptr=%lx\n", (unsigned long)buf->ptr); */
    buf->len = INITLEN;
    buf->gaplen = INITLEN;

    buf->head = mark_new(buf, 0, 0, 1, type | INSERT_AFTER);
    buf->tail = mark_new_default(buf, 0, 0, 1, type);
    buf->point = mark_new_default(buf, 0, 0, 1, type);

    return buf;
}

static void
buf_free(buf)
    struct textbuf *buf;
{
    struct bufmark *m;

    if (buf->ptr) free(buf->ptr);
    buf->ptr = 0;
    buf->len = 0;
    buf->gap = 0;

    remove_mark(buf->point);
    free(buf->point);
    buf->point = 0;
    for (m = buf->marks; m; m = m->next)
        m->flags &= ~ALIVE_FLAG;
    buf->marks = 0;
    
    free(buf);
}


static struct bufmark*
mark_new(buf, idx, cidx, line, flags)
    struct textbuf *buf;
    long idx, cidx, line;
    unsigned long flags;
{
    struct bufmark *mark;

    mark = ALLOC(struct bufmark);
    mark->buf  = buf;
    mark->idx  = idx;
    mark->cidx = cidx;
    mark->line = line;
    mark->prev = 0;
    mark->next = 0;

    mark->flags = ALIVE_FLAG | flags;

    mark->next = buf->marks;
    buf->marks = mark;
    adjust_mark_forward(mark);

    return mark;
}

static void
mark_free(mark)
    struct bufmark *mark;
{
    if (ALIVE_P(mark))
        remove_mark(mark);
    free(mark);
}



/*
==================================================================
                      mark list handle
==================================================================
*/

static void
insert_mark_between(mark, prev, next)
    struct bufmark *mark, *prev, *next;
{
    mark->prev = prev;
    mark->next = next;
    if (prev) prev->next = mark;
    if (next) next->prev = mark;
}

static void
remove_mark(mark)
    struct bufmark *mark;
{
    if (mark->prev)
        mark->prev->next = mark->next;
    if (mark->next)
        mark->next->prev = mark->prev;
    mark->prev = 0;
    mark->next = 0;
}

static struct bufmark*
first_mark(mark)
    struct bufmark *mark;
{
    struct bufmark *top;

    top = mark->buf->marks ? mark->buf->marks : mark;
    while (top->prev)
        top = top->prev;
    
    return top;
}

static void
adjust_mark_forward(mark)
    struct bufmark *mark;
{
    struct bufmark *prev, *next;

    prev = mark->prev;
    next = mark->next;
    remove_mark(mark);

    while (next) {
        if (mark->idx == next->idx) {
            if ((mark->flags & INSERT_BEFORE) <=
                (next->flags & INSERT_BEFORE))
                /* ins-after is "little" than ins-before */
                break;
        }
        else if (mark->idx < next->idx) {
            /* if ((prev && prev->idx < mark->idx) || !prev) */
            break;
        }
        prev = next;
        next = next->next;
    }
    insert_mark_between(mark, prev, next);

    mark->buf->marks = first_mark(mark);
}

static void
adjust_mark_backward(mark)
    struct bufmark *mark;
{
    struct bufmark *top, *prev, *next;

    prev = mark->prev;
    next = mark->next;
    remove_mark(mark);

    while (prev) {
        if (prev->idx == mark->idx) {
            if ((prev->flags & INSERT_BEFORE) <=
                (mark->flags & INSERT_BEFORE))
                /* ins-after is "little" than ins-before */
                break;
        }
        else if (prev->idx < mark->idx) {
            /* if ((next && mark->idx < next->idx) || !next) */
            break;
        }
        next = prev;
        prev = prev->prev;
    }

    insert_mark_between(mark, prev, next);

    top = mark->buf->marks ? mark->buf->marks : mark;
    while (top->prev)
        top = top->prev;
    mark->buf->marks = top;
}



/*
===================================================================
                        mark handle
===================================================================
*/

#define linehead_m(m) linehead(m->buf, m->idx)
#define linetail_m(m) linetail(m->buf, m->idx)

static void
mark_forward_char(mark, clen)
    struct bufmark *mark;
    long clen;
{
    struct bufmark *m;
    long targ;
    long line;

    targ = legal_cidx(mark->buf, mark->cidx + clen);
    for (m = mark; m->next; m = m->next) {
        if (m->next->cidx > targ)
            break;
    }
    mark->idx = byte_idx(mark->buf, m->idx, targ - m->cidx, &line);
    mark->cidx = targ;
    mark->line += line;

    adjust_mark_forward(mark);
}

static void
mark_forward_byte(mark, len)
    struct bufmark *mark;
    long len;
{
    long line;

    len = legal_len(mark->buf, mark->idx, len);
    mark->cidx += char_length(mark->buf, mark->idx, len, &line);
    mark->idx += len;
    mark->line += line;

    adjust_mark_forward(mark);
}


static void
mark_back_char(mark, clen)
    struct bufmark *mark;
    long clen;
{
    struct bufmark *m;
    long targ, i, cl, line;

    targ = legal_cidx(mark->buf, mark->cidx - clen);
    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->cidx < targ) {
            m = m->prev;
            break;
        }
    }
    if (m->cidx < targ) {
        i = m->idx;
        cl = targ - m->cidx;
    }
    else {
        m = 0;
        i = 0;
        cl = targ;
    }
    mark->idx = byte_idx(mark->buf, i, cl, &line);
    mark->cidx = targ;
    mark->line = m ? m->line + line : 1 + line;

    adjust_mark_backward(mark);
}

static void
mark_back_byte(mark, len)
    struct bufmark *mark;
    long len;
{
    long line;

/* if (dodebug) printf("back_byte: len=%ld\n", len); */
    len = legal_len(mark->buf, mark->idx, -len);
    len = -len;
/* if (dodebug) printf("back_byte: len=%ld\n", len); */
    mark->idx -= len;
/* if (dodebug) printf("back_byte: mark->idx=%ld, len=%ld\n",
                       mark->idx, len); */
    mark->cidx -= char_length(mark->buf, mark->idx, len, &line);
/* if (dodebug) printf("back_byte: mark->line %ld, line %ld\n",
                       mark->line, line); */
    mark->line -= line;

    adjust_mark_backward(mark);
}

static void
mark_move(mark, len)
    struct bufmark *mark;
    long len;
{
    if (len >= 0)
        mark_forward_char(mark, len);
    else
        mark_back_char(mark, -len);
}

static void
mark_nextline(mark)
    struct bufmark *mark;
{
    long i;

    i = linetail_m(mark);
    if (i < mark->buf->len) {
        i = fwdnl(mark->buf, i);
    }
    mark_forward_byte(mark, i - mark->idx);
}

static void
mark_prevline(mark)
    struct bufmark *mark;
{
    long i;

/* if (dodebug) { printf("prevline "); p_mark(mark); } */
    i = linehead_m(mark);
/* if (dodebug) printf("i(1) %ld\n", i); */
    if (i > 0) {
        i = backnl(mark->buf, i);
/* if (dodebug) printf("i(2) %ld\n", i); */
    }
    mark_back_byte(mark, mark->idx - i);
}


static void
mark_forward_nline(mark, n)
    struct bufmark *mark;
    long n;
{
    long i;

    i = forward_nline(mark->buf, mark->idx, n);
    mark_forward_byte(mark, i - mark->idx);
}

static void
mark_back_nline(mark, n)
    struct bufmark *mark;
    long n;
{
    long i;

    i = back_nline(mark->buf, mark->idx, n);
    mark_back_byte(mark, mark->idx - i);
}


static void
mark_linetail(mark)
    struct bufmark *mark;
{
    long i;

    i = linetail_m(mark);
    mark_forward_byte(mark, i - mark->idx);
}

static void
mark_linehead(mark)
    struct bufmark *mark;
{
    long i;

    i = linehead_m(mark);
    mark_back_byte(mark, mark->idx - i);
}


static long
mark_col(mark)
    struct bufmark *mark;
{
    long i;

    i = linehead_m(mark);
    return char_length(mark->buf, i, mark->idx - i, 0);
}

static long
mark_bcol(mark)
    struct bufmark *mark;
{
    return mark->idx - linehead_m(mark);
}

static void
mark_setcol(mark, i)
    struct bufmark *mark;
    long i;
{
    long head, orig;

    orig = mark->idx;
    head = linehead_m(mark);
    i = fwdcol(mark->buf, head, i);
    mark->idx = i;
    mark->cidx = char_idx(mark->buf, head, i - head, 0);
    if (orig > i)
        adjust_mark_backward(mark);
    else
        adjust_mark_forward(mark);
}


static void
mark_insert(mark, str, len)
    struct bufmark *mark;
    char *str;
    long len;
{
    struct bufmark *m;
    long clen, line;
    long mi = mark->idx;   /* We must be save this because mark->idx
                              will be changed in loop */

    insert(mark->buf, mark->idx, str, len);

    clen = char_length(mark->buf, mark->idx, len, &line);

    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != m->idx)
            break;
    }
    for (; m && m->idx == mi; m = m->next) {
        if (m->flags & INSERT_BEFORE) {
            m->idx += len;
            m->cidx += clen;
            m->line += line;
        }
    }
    for (; m; m = m->next) {
        m->idx += len;
        m->cidx += clen;
        m->line += line;
    }
}

static void
mark_delete(mark, len)
    struct bufmark *mark;
    long len;
{
    struct bufmark *m;
    long upper, clen, line;

    clen = char_length(mark->buf, mark->idx, len, &line);
    delete(mark->buf, mark->idx, len);

    upper = mark->idx + len;

    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != m->idx)
            break;
    }
    for (; m && m->idx < upper; m = m->next) {
        m->idx = mark->idx;
        m->cidx = mark->cidx;
        m->line = mark->line;
    }
    for (; m; m = m->next) {
        m->idx -= len;
        m->cidx -= clen;
        m->line -= line;
    }
}



/*
===================================================================
                    mark handle (strbuf)
===================================================================
*/

static void
smark_move(mark, len)
    struct bufmark *mark;
    long len;
{
    len = legal_len(mark->buf, mark->idx, len);
    if (len > 0) {
        mark->idx += len;
        adjust_mark_forward(mark);
    }
    else {
        mark->idx += len;
        adjust_mark_backward(mark);
    }
}

static void
smark_insert(mark, str, len)
    struct bufmark *mark;
    char *str;
    long len;
{
    struct bufmark *m;
    long mi = mark->idx;   /* We must be save this because mark->idx
                              will be changed in loop */

    insert(mark->buf, mark->idx, str, len);

    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != m->idx)
            break;
    }
    for (; m && m->idx == mi; m = m->next) {
        if (m->flags & INSERT_BEFORE)
            m->idx += len;
    }
    for (; m; m = m->next) {
        m->idx += len;
    }
}

static void
smark_delete(mark, len)
    struct bufmark *mark;
    long len;
{
    struct bufmark *m;
    long upper, beg;

    beg = mark->idx;
    len = legal_len(mark->buf, mark->idx, len);
    if (len < 0) {
        beg += len;
        len = -len;
    }
    delete(mark->buf, beg, len);

    upper = beg + len;
    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != beg)
            break;
    }
    for (; m && m->idx < upper; m = m->next) {
        m->idx = beg;
    }
    for (; m; m = m->next) {
        m->idx -= len;
    }
}


/*
===================================================================
                     byte string handle
===================================================================
*/

#define REAL_INDEX(buf,i) ((i >= buf->gap) ? i + buf->gaplen : i)
#define PSEDO_INDEX(buf,i) ((i >= buf->gap) ? i - buf->gaplen : i)

#define TB_EACH_CHAR(buf,beg,var) \
do {                                                              \
    long edge = beg >= buf->gap ? buf->len : buf->gap;            \
    var = REAL_INDEX(buf, beg);                                   \
  ec_again:                                                       \
    while (var < edge) {

#define TB_EACH_CHAR_END(buf,var) \
    }                                                             \
    if (edge == buf->gap) {                                       \
        var += buf->gaplen;                                       \
        edge = buf->len;                                          \
        goto ec_again;                                            \
    }                                                             \
  ec_end:                                                         \
    if (var > buf->gap)                                           \
        var -= buf->gaplen;                                       \
} while (0);

#define TB_EACH_CHAR_REVERSE(buf,beg,var) \
do {                                                              \
    long edge = beg >= buf->gap ? buf->gap + buf->gaplen : 0;      \
    var = REAL_INDEX(buf, beg);                                   \
  ec_again:                                                       \
    while (var > edge) {

#define TB_EACH_CHAR_REVERSE_END(buf,var) \
    }                                                             \
    if (edge != 0) {                                              \
        var -= buf->gaplen;                                       \
        edge = 0;                                                 \
        goto ec_again;                                            \
    }                                                             \
  ec_end:                                                         \
      if (var > buf->gap)                                         \
        var -= buf->gaplen;                                       \
} while (0);

#define TB_BREAK(buf,var) goto ec_end


#define FWDNL(buf, i) do { \
    long edge;                                                            \
    if (i < buf->gap) {                                                   \
        edge = buf->gap;                                                  \
    } else if (i == buf->gap) {                                           \
        i += buf->gaplen;                                                 \
        edge = buf->len;                                                  \
    } else {                                                              \
        edge = buf->len;                                                  \
    }                                                                     \
    if (i != edge) {                                                      \
        i++;                                                              \
        if (buf->ptr[i-1] == '\r' && i != edge && buf->ptr[i] == '\n')    \
            i++;                                                          \
    }                                                                     \
} while (0)

#define BACKNL(buf, i) do { \
    long edge;                                                            \
    if (i == buf->gap + buf->gaplen) {                                    \
        i -= buf->gaplen;                                                 \
        edge = 0;                                                         \
    } else if (i > buf->gap) {                                            \
        edge = buf->gap + buf->gaplen;                                    \
    } else {                                                              \
        edge = 0;                                                         \
    }                                                                     \
    if (i != edge) {                                                      \
        i--;                                                              \
        if (buf->ptr[i] == '\n' && i != edge && buf->ptr[i-1] == '\r')    \
            i--;                                                          \
    }                                                                     \
} while (0)


/* must be called idx is just before NL */
static long
fwdnl(buf, idx)
    struct textbuf *buf;
    long idx;
{
    idx = REAL_INDEX(buf, idx);
    FWDNL(buf, idx);
    return PSEDO_INDEX(buf, idx);
}

/* must be called idx is just after NL */
static long
backnl(buf, idx)
    struct textbuf *buf;
    long idx;
{
    idx = REAL_INDEX(buf, idx);
    BACKNL(buf, idx);
    return PSEDO_INDEX(buf, idx);
}


#if 0
/* ?.... */
static long
searchi(buf, beg, argtype, arg, ch, li)
    struct textbuf *buf;
    int argtype;
    long beg, arg;
    long *ch, *li;
{
    long i, targ, clen, line, d1, d2;

    targ = clen = line = 0;
    if (argtype & LEN_STOP) {
        targ = beg + arg;
        targ = REAL_INDEX(buf, targ);
    }
    if (argtype & CLEN_STOP) {
        clen = arg;
    }
    if (argtype & LINE_STOP) {
        line = arg;
    }
    if (!ch) ch = &d1; *ch = 0;
    if (!li) li = &d2; *li = 0;
    
    TB_EACH_CHAR(buf, beg, i)
        if ((argtype & LEN_STOP) && i >= targ)
            TB_BREAK(buf, i);
        if ((argtype & CLEN_STOP) && !clen)
            TB_BREAK(buf, i);
        switch (buf->ptr[i]) {
        case '\n':
        case '\r':
            if (argtype & NO_LINE)
                TB_BREAK(buf, i);
            FWDNL(buf, i);
            line--;
            (*li)++;
            if ((argtype & LINE_STOP) && !line)
                TB_BREAK(buf, i);
            break;
        default:
            i += mbclen(buf->ptr[i]);
            break;
        }
        clen--;
        (*ch)++;
    TB_EACH_CHAR_END(buf, i)
    return i;
}
#endif


static long
linehead(buf, idx)
    struct textbuf *buf;
    long idx;
{
    long i;
    
    TB_EACH_CHAR_REVERSE(buf, idx, i)
        if (buf->ptr[i-1] == '\n' || buf->ptr[i-1] == '\r')
            TB_BREAK(buf, i);
        i--;
    TB_EACH_CHAR_REVERSE_END(buf, i)
    return i;
}

static long
linetail(buf, idx)
    struct textbuf *buf;
    long idx;
{
    long i;

    TB_EACH_CHAR(buf, idx, i)
        if (buf->ptr[i] == '\n' || buf->ptr[i] == '\r')
            TB_BREAK(buf, i);
        i++;
    TB_EACH_CHAR_END(buf, i)
    return i;
}


static long
forward_nline(buf, idx, n)
    struct textbuf *buf;
    long idx, n;
{
    long i;

    if (!n) return idx;
    TB_EACH_CHAR(buf, idx, i)
        if (buf->ptr[i] == '\n' || buf->ptr[i] == '\r') {
            FWDNL(buf, i);
            n--;
            if (!n)
                TB_BREAK(buf, i);
        }
        else {
            i++;
        }
    TB_EACH_CHAR_END(buf, i)
    return i;
}

static long
back_nline(buf, idx, n)
    struct textbuf *buf;
    long idx, n;
{
    long i;

    if (!n) return idx;
    TB_EACH_CHAR_REVERSE(buf, idx, i)
        if (buf->ptr[i-1] == '\n' || buf->ptr[i-1] == '\r') {
            BACKNL(buf, i);
            n--;
            if (!n)
                TB_BREAK(buf, i);
        }
        else {
            i--;
        }
    TB_EACH_CHAR_REVERSE_END(buf, i)
    return i;
}


static long
byte_idx(buf, begidx, clen, line)
    struct textbuf *buf;
    long begidx, clen;
    long *line;
{
    long i;
    long c = clen;
    long dummy;

    Db(printf("byte_idx> beg=%ld, clen=%ld\n", begidx, clen));
    Db(p_buf(buf));

    if (!line) line = &dummy;
    (*line) = 0;
    TB_EACH_CHAR(buf, begidx, i)
        if (!c)
            TB_BREAK(buf, i);
        switch (buf->ptr[i]) {
        case '\n':
        case '\r':
            FWDNL(buf, i);
            (*line)++;
            break;
        default:
            i += mbclen(buf->ptr[i]);
            break;
        }
        c--;
    TB_EACH_CHAR_END(buf, i)

    Db(printf("byte_idx< clen/len=%ld/%ld\n", clen, i - begidx));
    return i;
}

static long
byte_length(buf, begidx, clen, line)
    struct textbuf *buf;
    long begidx, clen;
    long *line;
{
    return byte_idx(buf, begidx, clen, line) - begidx;
}


static long
char_length(buf, begidx, len, line)
    struct textbuf *buf;
    long begidx, len;
    long *line;
{
    long i;
    long c = 0;
    long targ;
    long dummy;

    Dc(printf("char_len> beg=%ld, len=%ld\n", begidx, len));
    Dc(p_buf(buf));

    if (!line) line = &dummy;
    (*line) = 0;
    targ = legal_idx(buf, begidx + len);
    targ = REAL_INDEX(buf, targ);

    TB_EACH_CHAR(buf, begidx, i)
        if (i == targ)
            TB_BREAK(buf, i);
/* if (dodebug) { printf("%x", buf->ptr[i]); fflush(stdout); } */
        switch (buf->ptr[i]) {
        case '\n':
        case '\r':
            FWDNL(buf, i);
            (*line)++;
            break;
        default:
            i += mbclen(buf->ptr[i]);
            break;
        }
        c++;
    TB_EACH_CHAR_END(buf, i)

    Dc(printf("char_len< clen=%ld line=%+ld\n", c, *line));
    return c;
}

static long
char_idx(buf, begidx, len, line)
    struct textbuf *buf;
    long begidx, len;
    long *line;
{
    return begidx + char_length(buf, begidx, len, line);
}


static long
fwdcol(buf, beg, clen)
    struct textbuf *buf;
    long beg, clen;
{
    long i;
    long c = clen;

    Db(printf("fwdcol> beg=%ld, clen=%ld\n", beg, clen));
    Db(p_buf(buf));

    TB_EACH_CHAR(buf, beg, i)
        if (!c)
            TB_BREAK(buf, i);
        switch (buf->ptr[i]) {
        case '\n':
        case '\r':
            TB_BREAK(buf, i);
            break;
        default:
            i += mbclen(buf->ptr[i]);
            break;
        }
        c--;
    TB_EACH_CHAR_END(buf, i)

    Db(printf("fwdcol< clen/len=%ld/%ld\n", clen, i - beg));
    return i;
}


static void
setgap(buf, idx)
    struct textbuf *buf;
    long idx;
{
    if (idx < buf->gap) {
        memmove(buf->ptr + (idx + buf->gaplen),
                buf->ptr + idx,
                buf->gap - idx);
    }
    else {
        memmove(buf->ptr + buf->gap,
                buf->ptr + (buf->gap + buf->gaplen),
                idx - buf->gap);
    }
    buf->gap = idx;
}


#define ISIZE 1024

static void
insert(buf, idx, str, len)
    struct textbuf *buf;
    long idx;
    char *str;
    long len;
{
    if (buf->gaplen < len) {
        long newcapa, inc;
        
        if (len < ISIZE)
            inc = ISIZE;
        else
            inc = len * 2;
        newcapa = buf->len + inc;
        REALLOC_N(buf->ptr, char, newcapa);

        if (idx < buf->gap) {
            long pl1, pl2;
            pl1 = buf->len - buf->gap - buf->gaplen;
            memmove(buf->ptr + (newcapa - pl1),
                    buf->ptr + (buf->gap + buf->gaplen),
                    pl1);
            pl2 = buf->gap - idx;
            memmove(buf->ptr + (newcapa - pl1 - pl2),
                    buf->ptr + idx,
                    pl2);
        }
        else {
            long pl1, pl2;
            pl1 = buf->len - (idx + buf->gaplen);
            memmove(buf->ptr + (newcapa - pl1),
                    buf->ptr + (buf->len - pl1),
                    pl1);
            pl2 = idx - buf->gap;
            memmove(buf->ptr + buf->gap,
                    buf->ptr + buf->gap + buf->gaplen,
                    pl2);
        }
        buf->len = newcapa;
        buf->gap = idx;
        buf->gaplen += inc;
    }
    else {
        setgap(buf, idx);
    }
    memmove(buf->ptr + idx,
            str,
            sizeof(char) * len);
    buf->gap += len;
    buf->gaplen -= len;
}


static void
delete(buf, idx, len)
    struct textbuf *buf;
    long idx, len;
{
    Dd(printf("\ndelete\n"));
    Dd(p_buf(buf));
    Dd(printf("idx=%ld, len=%ld\n", idx, len));

    if (idx < buf->gap) {
        if (idx + len > buf->gap) {
            /*
                 OOOOOxxxx-----xxOOOO

                 OOOOO-----------OOOO
            */
            buf->gap = idx;
            buf->gaplen += len;
        }
        else {
            /*
                 OOOOOxxxxxxXXX----OOOOOO

                 OOOOO----------XXXOOOOOO
            */
            memmove(buf->ptr + (idx + len) + buf->gaplen,
                    buf->ptr + (idx + len),
                    buf->gap - (idx + len));
            buf->gap = idx;
            buf->gaplen += len;
        }
    }
    else {
        /*
             OOOOO---XXxxxOOOO

             OOOOOXX------OOOO
        */
        long pl = idx - buf->gap;

        memmove(buf->ptr + buf->gap,
                buf->ptr + buf->gap + buf->gaplen,
                pl);
        buf->gap += pl;
        buf->gaplen += len;
    }
}


static VALUE
substr(buf, idx, len)
    struct textbuf *buf;
    long idx, len;
{
    Ds(printf("substr: idx=%ld, len=%ld\n", idx, len));
    if ((idx <= buf->gap) && (idx + len > buf->gap)) {
        VALUE ret;
        Ds(printf("1=%ld 2=%ld\n", buf->gap - idx, idx + len - buf->gap));
        ret = rb_str_new(buf->ptr + idx, buf->gap - idx);
        rb_str_cat(ret, buf->ptr + buf->gap + buf->gaplen,
                        idx + len - buf->gap);
        return ret;
    }
    else {
        Ds(printf("1=%ld\n", len));
        return rb_str_new(buf->ptr + REAL_INDEX(buf, idx), len);
    }
}


static VALUE
getline(buf, idx)
    struct textbuf *buf;
    long idx;
{
    long head, tail;

    head = linehead(buf, idx);
    tail = linetail(buf, idx);
    if (head == LEN(buf))
        return Qnil;
    return substr(buf, head, tail - head);
}


struct roeach_param {
    struct textbuf *buf;
    long idx;
    struct RString *str;
};

static VALUE
roeach_i(arg)
    VALUE arg;
{
    struct roeach_param *p = (struct roeach_param*)arg;
    struct textbuf *buf = p->buf;
    struct RString *str = p->str;
    long head, tail;

    head = p->idx;
    while (head < LEN(buf)) {
        tail = linetail(buf, head);
        if (head < buf->gap && tail > buf->gap)
            setgap(buf, tail);
        str->ptr = buf->ptr + head;
        str->len = tail - head;
        rb_yield((VALUE)str);
        head = fwdnl(buf, tail);
    }
    return Qnil;
}

static VALUE
roeach_ensure(arg)
    VALUE arg;
{
    struct roeach_param *p = (struct roeach_param*)arg;

    RSTRING(p->str)->ptr = 0;
    RSTRING(p->str)->len = 0;
    return Qnil;
}

static void
each_line_ro(buf, idx)
    struct textbuf *buf;
    long idx;
{
    struct roeach_param p;

    NEWOBJ(str, struct RString);
    OBJSETUP(str, rb_cString, T_STRING);
    str->ptr = 0;
    str->len = 0;
    str->orig = 0;
    OBJ_FREEZE(str);

    p.str = str;
    p.buf = buf;
    p.idx = idx;
    rb_ensure(roeach_i, (VALUE)&p, roeach_ensure, (VALUE)&p);
}


static long
legal_cidx(buf, c)
    struct textbuf *buf;
    long c;
{
    if (c < 0) return 0;
    if (c > CLEN(buf)) return CLEN(buf);
    return  c;
}

static long
legal_clen(buf, i, clen)
    struct textbuf *buf;
    long i, clen;
{
    if (i + clen > CLEN(buf)) return CLEN(buf) - i;
    return clen;
}

static long
legal_idx(buf, i)
    struct textbuf *buf;
    long i;
{
    if (i < 0) return 0;
    if (i > LEN(buf)) return LEN(buf);
    return i;
}

static long
legal_len(buf, i, len)
    struct textbuf *buf;
    long i, len;
{
    if (i + len > LEN(buf)) return LEN(buf) - i;
    if (i + len < 0) return -i;
    return len;
}



/*
########################################################################
########################################################################
                        ruby interface  
########################################################################
########################################################################
*/

#define GET_BUF(v,var) Data_Get_Struct(v, struct textbuf, var)
#define GET_MARK(v, var) Data_Get_Struct(v, struct bufmark, var)
#define GET_MARK_CHK(v, var) GET_MARK(v, var); alive_check(var)

#define WRAP_BUF(p, c) Data_Wrap_Struct(c, buffer_mark, buffer_free, p)
#define WRAP_MARK(p, c) Data_Wrap_Struct(c, 0, bufmark_free, p)

/*
========================================================================
                           Buffer
========================================================================
*/

static void
buffer_mark(p)
    void *p;
{
    struct textbuf *buf = (struct textbuf*)p;
    rb_gc_mark(buf->vhead);
    rb_gc_mark(buf->vtail);
}

static void
buffer_free(p)
    void *p;
{
    struct textbuf *buf = (struct textbuf*)p;
    buf->vhead = Qnil;
    buf->vtail = Qnil;
    buf_free(buf);
}

static void
bufmark_free(p)
    void *p;
{
    mark_free((struct bufmark*)p);
}

static VALUE
buffer_s_new(self)
    VALUE self;
{
    struct textbuf *buf;
    volatile VALUE v[2];
    VALUE markc;
    long flag;

    if (self == TextBuffer) {
        markc = TextBufferMark;
        flag = TEXTBUFMARK_FLAG;
    }
    else {
        markc = StringBufferMark;
        flag = STRBUFMARK_FLAG;
    }
    buf = buf_new(flag);
    buf->vhead = v[0] = WRAP_MARK(buf->head, markc);
    OBJ_FREEZE(buf->vhead);
    buf->vtail = v[1] = WRAP_MARK(buf->tail, markc);
    OBJ_FREEZE(buf->vtail);

    return WRAP_BUF(buf, self);
}

static VALUE
buffer_new_mark(argc, argv, self)
    int argc;
    VALUE *argv;
    VALUE self;
{
    struct textbuf *buf;
    struct bufmark *mark;
    long c = 0;
    VALUE markc;

    switch (argc) {
    case 0:
        c = 0;
        break;
    case 1:
        c = NUM2LONG(*argv);
        break;
    default:
        rb_raise(rb_eArgError, "too many arguments (%d for 0/1)", argc);
        break;
    }

    GET_BUF(self, buf);
    if (CLASS_OF(self) == TextBuffer) {
        long line, i;
        c = legal_cidx(buf, c);
        i = byte_length(buf, 0, c, &line);
        mark = mark_new_default(buf, i, c, 1 + line, TEXTBUFMARK_FLAG);
        markc = TextBufferMark;
    }
    else {
        mark = mark_new_default(buf, legal_idx(buf, c), 0, 0, STRBUFMARK_FLAG);
        markc = StringBufferMark;
    }

    return WRAP_MARK(mark, markc);
}


static VALUE
buffer_head(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return buf->vhead;
}

static VALUE
buffer_tail(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return buf->vtail;
}


static VALUE
buffer_char_size(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return INT2NUM(CLEN(buf));
}

static VALUE
buffer_byte_size(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return INT2NUM(LEN(buf));
}


static VALUE
buffer_concat(self, str)
    VALUE self, str;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    if (TEXTBUFMARK_P(buf->tail))
        mark_insert(buf->tail, RSTRING(str)->ptr, RSTRING(str)->len);
    else
        smark_insert(buf->tail, RSTRING(str)->ptr, RSTRING(str)->len);

    return self;
}


static VALUE
buffer_string(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return substr(buf, 0, LEN(buf));
}



/*
========================================================================
                          BufferMark
========================================================================
*/

#define bufmark_p(v) \
    ((TYPE(v) == T_DATA) && (RDATA(v)->dfree == bufmark_free))

#define origin_check(m1, m2) \
    if (m1->buf != m2->buf) \
        rb_raise(rb_eArgError, "marks belonging to different buffer used"); \
    else

#define freeze_check(m) \
    if (OBJ_FROZEN(m)) \
        rb_raise(rb_eArgError, "can't move frozen mark"); \
    else

#define alive_check(mark) \
    if (!ALIVE_P(mark)) \
        rb_raise(rb_eArgError, "method called for dead mark"); \
    else

static VALUE
bufmark_dup(self)
    VALUE self;
{
    struct bufmark *mark, *p;
    VALUE clone;

    GET_MARK_CHK(self, mark);
    p = mark_new_default(mark->buf, mark->idx, mark->cidx, mark->line,
                         MARKTYPE(mark));
    clone = WRAP_MARK(p, CLASS_OF(self));
    CLONESETUP(clone, self);

    return clone;
}

static VALUE
bufmark_clone(self)
    VALUE self;
{
    VALUE tmp = bufmark_dup(self);
    if (OBJ_FROZEN(self)) OBJ_FREEZE(tmp);
    return tmp;
}


static VALUE
bufmark_alive_p(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    return ALIVE_P(mark) ? Qtrue : Qfalse;
}

static VALUE
bufmark_discard(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    remove_mark(mark);
    mark->flags &= ~ALIVE_FLAG;

    return Qnil;
}


static VALUE
bufmark_insert(self, str)
    VALUE self, str;
{
    struct bufmark *mark;
    
    GET_MARK_CHK(self, mark);
    if (TEXTBUFMARK_P(mark))
        mark_insert(mark, RSTRING(str)->ptr, RSTRING(str)->len);
    else
        smark_insert(mark, RSTRING(str)->ptr, RSTRING(str)->len);

    return self;
}

static void sbm_beg_len _((struct bufmark*, VALUE, long*, long*));
static void tbm_begm_len _((struct bufmark*, VALUE, struct bufmark**, long*));

static VALUE
bufmark_copyof(self, i, l)
    VALUE self, i, l;
{
    struct bufmark *mark, *m;
    struct textbuf *buf;
    long beg, len;
    void (*func)(struct bufmark*, char*, long);

    GET_MARK_CHK(self, mark);
    if (! bufmark_p(i))
        rb_raise(rb_eTypeError, "wrong arg type %s (expected BufferMark)",
                 rb_class2name(CLASS_OF(i)));
    GET_MARK_CHK(i, m);
    buf = m->buf;

    if (TEXTBUFMARK_P(m)) {
        struct bufmark *begm;

        tbm_begm_len(m, l, &begm, &len);
        beg = begm->idx;
        func = mark_insert;
    }
    else {
        sbm_beg_len(m, l, &beg, &len);
        func = smark_insert;
    }
    if (beg <= buf->gap && beg + len > buf->gap) {
        (*func)(mark,
                buf->ptr + REAL_INDEX(buf, beg),
                buf->gap - beg);
        (*func)(mark,
                buf->ptr + REAL_INDEX(buf, buf->gap),
                beg + len - buf->gap);
    }
    else {
        (*func)(mark, buf->ptr + REAL_INDEX(buf, beg), len);
    }

    return self;
}

static VALUE
bufmark_roeach(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    each_line_ro(mark->buf, mark->idx);

    return self;
}


static VALUE
bufmark_fwd(self, len)
    VALUE self, len;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    if (TEXTBUFMARK_P(mark))
        mark_move(mark, NUM2LONG(len));
    else
        smark_move(mark, NUM2LONG(len));

    return self;
}

static VALUE
bufmark_back(self, v)
    VALUE self, v;
{
    struct bufmark *mark;
    long len;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    len = NUM2LONG(v);
    if (TEXTBUFMARK_P(mark))
        mark_move(mark, -len);
    else
        smark_move(mark, -len);

    return self;
}


static VALUE
bufmark_char_index(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    return INT2NUM(mark->cidx);
}

static VALUE
bufmark_byte_index(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    return INT2NUM(mark->idx);
}

static VALUE
bufmark_setidx(self, idx)
    VALUE self;
{
    struct bufmark *mark;
    long i = 0;

    GET_MARK_CHK(self, mark);
    freeze_check(self);

    if (FIXNUM_P(idx)) {
        i = FIX2LONG(idx);
    }
    else if (bufmark_p(idx)) {
        struct bufmark *m;
        GET_MARK_CHK(idx, m);
        origin_check(mark, m);
        i = TEXTBUFMARK_P(m) ? m->cidx : m->idx;
    }
    else {
        i = NUM2LONG(idx);
    }

    if (TEXTBUFMARK_P(mark))
        mark_move(mark, i);
    else
        smark_move(mark, i);

    return self;
}


static VALUE
bufmark_go_beg(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark->idx = 0;
    if (TEXTBUFMARK_P(mark)) {
        mark->cidx = 0;
        mark->line = 1;
    }
    adjust_mark_backward(mark);

    return self;
}

static VALUE
bufmark_go_end(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark->idx = LEN(mark->buf);
    if (TEXTBUFMARK_P(mark)) {
        mark->cidx = CLEN(mark->buf);
        mark->line = mark->buf->tail->line;
    }
    adjust_mark_forward(mark);

    return self;
}


static VALUE
bufmark_plus(self, idx)
    VALUE self, idx;
{
    struct bufmark *mark, *m;
    VALUE ret;

    GET_MARK_CHK(self, mark);
    ret = bufmark_dup(self);
    GET_MARK(ret, m);
    if (TEXTBUFMARK_P(mark))
        mark_move(m, NUM2LONG(idx));
    else
        smark_move(m, NUM2LONG(idx));

    return ret;
}

static VALUE
bufmark_minus(self, idx)
    VALUE self, idx;
{
    struct bufmark *mark, *m;
    VALUE ret;
    long i;

    GET_MARK_CHK(self, mark);
    ret = bufmark_dup(self);
    GET_MARK(ret, m);
    i = NUM2LONG(idx);
    if (TEXTBUFMARK_P(mark))
        mark_move(m, -i);
    else
        smark_move(m, -i);

    return ret;
}



/*
========================================================================
                       class StringBuffer
========================================================================
*/

static VALUE
strbuf_substr(self, idx, len)
    VALUE self, idx, len;
{
    struct textbuf *buf;
    long i, l;

    GET_BUF(self, buf);
    i = legal_idx(buf, NUM2LONG(idx));
    l = legal_len(buf, i, NUM2LONG(len));

    return substr(buf, i, l);
}

static VALUE
strbuf_inspect(self)
    VALUE self;
{
    struct textbuf *buf;
    char tmp[128];
    int len;

    GET_BUF(self, buf);
    len = sprintf(tmp,
      "#<StringBuffer: 0x%lx len=%ld gap=%ld glen=%ld L%ld>",
      (unsigned long)buf,
      buf->len, buf->gap, buf->gaplen, LEN(buf));
    return rb_str_new(tmp, len);
}


/*
========================================================================
                     class StringBufferMark
========================================================================
*/

static void
sbm_beg_len(mark, val, beg, len)
    struct bufmark *mark;
    VALUE val;
    long *beg, *len;
{
    if (FIXNUM_P(val)) {
        *beg = mark->idx;
        *len = legal_len(mark->buf, *beg, FIX2LONG(val));
    }
    else if (bufmark_p(val)) {
        struct bufmark *m;
        GET_MARK_CHK(val, m);
        origin_check(mark, m);
        if (m->idx < mark->idx) {
            *beg = m->idx;
            *len = mark->idx - m->idx;
        }
        else {
            *beg = mark->idx;
            *len = m->idx - mark->idx;
        }
    }
    else {
        *len = NUM2LONG(val);
    }
}


static VALUE
sbufmark_delete(self, val)
    VALUE self, val;
{
    struct bufmark *mark;
    long beg, len;

    GET_MARK_CHK(self, mark);
    sbm_beg_len(mark, val, &beg, &len);
    smark_delete(mark, len);

    return self;
}


static VALUE
sbufmark_substr(self, val)
    VALUE self, val;
{
    struct bufmark *mark;
    long beg, len;

    GET_MARK_CHK(self, mark);
    sbm_beg_len(mark, val, &beg, &len);

    return substr(mark->buf, beg, len);
}


static VALUE
sbufmark_inspect(self)
    VALUE self;
{
    struct bufmark *mark;
    char buf[128];
    int len;

    GET_MARK(self, mark);
    if (ALIVE_P(mark)) {
        len = sprintf(buf, "#<StringBufferMark: 0x%lx buf=0x%lx i=%ld>",
                      (unsigned long)mark, (unsigned long)mark->buf,
                      mark->idx);
    }
    else {
        len = sprintf(buf, "#<StringBufferMark: 0x%lx dead>",
                      (unsigned long)mark);
    }
    return rb_str_new(buf, len);
}



/*
========================================================================
                       class TextBuffer
========================================================================
*/

static VALUE
textbuf_substr(self, cidx, clen)
    VALUE self, cidx, clen;
{
    struct textbuf *buf;
    long ci, len;

    GET_BUF(self, buf);
    ci = legal_cidx(buf, NUM2LONG(cidx));
    if (buf->point->cidx > ci)
        mark_back_char(buf->point, buf->point->cidx - ci);
    else
        mark_forward_char(buf->point, ci - buf->point->cidx);
    len = byte_length(buf, buf->point->idx, NUM2LONG(clen), 0);

    return substr(buf, buf->point->idx, len);
}

static VALUE
textbuf_inspect(self)
    VALUE self;
{
    struct textbuf *buf;
    char tmp[128];
    int len;

    GET_BUF(self, buf);
    len = sprintf(tmp,
      "#<TextBuffer: 0x%lx len=%ld gap=%ld glen=%ld L%ld C%ld>",
      (unsigned long)buf,
      buf->len, buf->gap, buf->gaplen, LEN(buf), CLEN(buf));
    return rb_str_new(tmp, len);
}



/*
========================================================================
                    class TextBufferMark
========================================================================
*/

static void
tbm_begm_len(mark, val, mvar, lvar)
    struct bufmark *mark;
    VALUE val;
    struct bufmark **mvar;
    long *lvar;
{
    long tmp;

    switch (TYPE(val)) {
    case T_FIXNUM:
        *mvar = mark;
        tmp = FIX2LONG(val);
        tmp = legal_clen(mark->buf, mark->cidx, tmp);
        *lvar = byte_length(mark->buf, mark->idx, tmp, 0);
        break;

    case T_DATA:
        if (bufmark_p(val)) {
            struct bufmark *m;
            GET_MARK_CHK(val, m);
            origin_check(mark, m);
            if (m->idx < mark->idx) {
                *mvar = m;
                *lvar = mark->idx - m->idx;
            }
            else {
                *mvar = mark;
                *lvar = m->idx - mark->idx;
            }
        }
        else {
            goto typeerr;
        }
        break;

    case T_BIGNUM:
        rb_raise(rb_eIndexError, "index too big");
        break;

    default:
        goto typeerr;
        break;
    }

    return;

typeerr:
    rb_raise(rb_eTypeError,
             "wrong arg type %s (expected Fixnum or TextBufferMark)",
             rb_class2name(CLASS_OF(val)));
}

static VALUE
tbufmark_delete(self, val)
    VALUE self, val;
{
    struct bufmark *mark, *m;
    long len;

    GET_MARK_CHK(self, mark);
    tbm_begm_len(mark, val, &m, &len);
    mark_delete(m, len);

    return self;
}

static VALUE
tbufmark_substr(self, val)
    VALUE self, val;
{
    struct bufmark *mark, *m;
    long len;

    GET_MARK_CHK(self, mark);
    tbm_begm_len(mark, val, &m, &len);

    return substr(mark->buf, m->idx, len);
}


static VALUE
tbufmark_line(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    return INT2NUM(mark->line);
}

static VALUE
tbufmark_setline(self, n)
    VALUE self;
{
    
    struct bufmark *mark;
    long i;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    i = NUM2LONG(n);
    if (mark->line > i) {
        mark_back_nline(mark, mark->line - i);
    }
    else {
        mark_forward_nline(mark, i - mark->line);
    }

    return n;
}

static VALUE
tbufmark_next_line(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark_nextline(mark);

    return self;
}

static VALUE
tbufmark_prev_line(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark_prevline(mark);

    return self;
}

static VALUE
tbufmark_line_head(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark_linehead(mark);

    return self;
}

static VALUE
tbufmark_line_tail(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    mark_linetail(mark);

    return self;
}

static VALUE
tbufmark_getline(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK_CHK(self, mark);

    return getline(mark->buf, mark->idx);
}

static VALUE
tbufmark_col(self)
    VALUE self;
{
    struct bufmark *mark;
    long i;

    GET_MARK_CHK(self, mark);
    i = mark_col(mark);
    return INT2NUM(i);
}

static VALUE
tbufmark_bytecol(self)
    VALUE self;
{
    struct bufmark *mark;
    long i;

    GET_MARK_CHK(self, mark);
    i = mark_bcol(mark);
    return INT2NUM(i);
}

static VALUE
tbufmark_setcol(self, n)
    VALUE self;
{
    
    struct bufmark *mark;
    long i;

    GET_MARK_CHK(self, mark);
    freeze_check(self);
    i = NUM2LONG(n);
    mark_setcol(mark, i);

    return n;
}


static VALUE
tbufmark_inspect(self)
    VALUE self;
{
    struct bufmark *mark;
    char buf[128];
    int len;

    GET_MARK(self, mark);
    if (ALIVE_P(mark)) {
        len = sprintf(buf, "#<TextBufferMark: 0x%lx buf=0x%lx i=%ld ci=%ld>",
                      (unsigned long)mark, (unsigned long)mark->buf,
                      mark->idx, mark->cidx);
    }
    else {
        len = sprintf(buf, "#<TextBufferMark: 0x%lx dead>",
                      (unsigned long)mark);
    }
    return rb_str_new(buf, len);
}



/*
========================================================================
                           init
========================================================================
*/

static VALUE
debugon(obj)
    VALUE obj;
{
    dodebug = 1; return Qtrue;
}

static VALUE
debugoff(obj)
    VALUE obj;
{
    dodebug = 0; return Qfalse;
}

void
Init_textbuf()
{
    StringBuffer = rb_define_class("StringBuffer", rb_cData);
    rb_define_singleton_method(StringBuffer, "new", buffer_s_new,    0);
    rb_define_method(StringBuffer, "new_mark",    buffer_new_mark,  -1);
    rb_define_method(StringBuffer, "head",        buffer_head,       0);
    rb_define_method(StringBuffer, "tail",        buffer_tail,       0);
    rb_define_method(StringBuffer, "size",        buffer_byte_size,  0);
    rb_define_method(StringBuffer, "length",      buffer_byte_size,  0);
    rb_define_method(StringBuffer, "byte_size",   buffer_byte_size,  0);
    rb_define_method(StringBuffer, "byte_length", buffer_byte_size,  0);
    rb_define_method(StringBuffer, "substr",      strbuf_substr,     2);
    rb_define_method(StringBuffer, "[]",          strbuf_substr,     2);
    rb_define_method(StringBuffer, "string",      buffer_string,     0);
    rb_define_method(StringBuffer, "to_s",        buffer_string,     0);
    rb_define_method(StringBuffer, "concat",      buffer_concat,     1);
    rb_define_method(StringBuffer, "<<",          buffer_concat,     1);
    rb_define_method(StringBuffer, "inspect",     strbuf_inspect,    0);

    StringBufferMark = rb_define_class("StringBufferMark", rb_cData);
    rb_define_method(StringBufferMark, "dup",        bufmark_dup,        0);
    rb_define_method(StringBufferMark, "clone",      bufmark_clone,      0);
    rb_define_method(StringBufferMark, "alive?",     bufmark_alive_p,    0);
    rb_define_method(StringBufferMark, "discard",    bufmark_discard,    0);
    rb_define_method(StringBufferMark, "insert",     bufmark_insert,     1);
    rb_define_method(StringBufferMark, "copy_of",    bufmark_copyof,     2);
    rb_define_method(StringBufferMark, "delete",     sbufmark_delete,    1);
    rb_define_method(StringBufferMark, "substr",     sbufmark_substr,    1);
    rb_define_method(StringBufferMark, "[]",         sbufmark_substr,    1);
    rb_define_method(StringBufferMark, "each_line_ro", bufmark_roeach,   0);
    rb_define_method(StringBufferMark, "index",      bufmark_byte_index, 0);
    rb_define_method(StringBufferMark, "byte_index", bufmark_byte_index, 0);
    rb_define_method(StringBufferMark, "to_i",       bufmark_byte_index, 0);
    rb_define_method(StringBufferMark, "index=",     bufmark_setidx,     0);
    rb_define_method(StringBufferMark, "move_to",    bufmark_setidx,     0);
    rb_define_method(StringBufferMark, "move",       bufmark_fwd,        1);
    rb_define_method(StringBufferMark, "forward",    bufmark_fwd,        1);
    rb_define_method(StringBufferMark, "back",       bufmark_back,       1);
    rb_define_method(StringBufferMark, "head",       bufmark_go_beg,     0);
    rb_define_method(StringBufferMark, "tail",       bufmark_go_end,     0);
    rb_define_method(StringBufferMark, "+",          bufmark_plus,       1);
    rb_define_method(StringBufferMark, "-",          bufmark_minus,      1);
    rb_define_method(StringBufferMark, "inspect",    sbufmark_inspect,   0);


    TextBuffer = rb_define_class("TextBuffer", StringBuffer);
    rb_define_method(TextBuffer, "size",        buffer_char_size, 0);
    rb_define_method(TextBuffer, "length",      buffer_char_size, 0);
    rb_define_method(TextBuffer, "[]",          textbuf_substr,   2);
    rb_define_method(TextBuffer, "inspect",     textbuf_inspect,  0);

    TextBufferMark = rb_define_class("TextBufferMark", StringBufferMark);
    rb_define_method(TextBufferMark, "delete",     tbufmark_delete,     1);
    rb_define_method(TextBufferMark, "substr",     tbufmark_substr,     1);
    rb_define_method(TextBufferMark, "[]",         tbufmark_substr,     1);
    rb_define_method(TextBufferMark, "getline",    tbufmark_getline,    0);
    rb_define_method(TextBufferMark, "index",      bufmark_char_index,  0);
    rb_define_method(TextBufferMark, "to_i",       bufmark_char_index,  0);
    rb_define_method(TextBufferMark, "line",       tbufmark_line,       0);
    rb_define_method(TextBufferMark, "y",          tbufmark_line,       0);
    rb_define_method(TextBufferMark, "line=",      tbufmark_setline,    1);
    rb_define_method(TextBufferMark, "y=",         tbufmark_setline,    1);
    rb_define_method(TextBufferMark, "column",     tbufmark_col,        0);
    rb_define_method(TextBufferMark, "x",          tbufmark_col,        0);
    rb_define_method(TextBufferMark, "column=",    tbufmark_setcol,     1);
    rb_define_method(TextBufferMark, "x=",         tbufmark_setcol,     1);
    rb_define_method(TextBufferMark, "byte_column",tbufmark_bytecol,    0);
    rb_define_method(TextBufferMark, "next_line",  tbufmark_next_line,  0);
    rb_define_method(TextBufferMark, "prev_line",  tbufmark_prev_line,  0);
    rb_define_method(TextBufferMark, "line_head",  tbufmark_line_head,  0);
    rb_define_method(TextBufferMark, "line_tail",  tbufmark_line_tail,  0);
    rb_define_method(TextBufferMark, "inspect",    tbufmark_inspect,    0);


    rb_define_method(rb_cObject, "textbuf_debug_on", debugon, 0);
    rb_define_method(rb_cObject, "textbuf_debug_off", debugoff, 0);
}
